# MIT License
#
# Copyright (c) 2021-2022 Templin Konstantin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Configuration model class module."""

from pprint import pformat


__all__ = ["Configuration"]


class Configuration(dict):
    """Configuration model, inherited from built-in dict.

    Example:
        .. code-block:: python

            from feijoa.model.configuration import Configuration

            configuration = Configuration(
                {"foo": 1, "bar": 2}, requestor="simple", request_id=0
            )

    Args:
        1st:
            configuration dict.
        requestor (str, optional):
            Name of requestor (algorithm).
        request_id (int, optional):
            Index of requested configuration
            for specified requestor.

    """

    def __init__(
        self,
        *args,
        requestor="UNKNOWN",
        request_id=0,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.requestor = requestor
        self.request_id = request_id

    def __str__(self):
        fmt = pformat(dict(self.items()))
        return (
            f"Configuration({fmt}, "
            f"requestor={self.requestor}, "
            f"request_id={self.request_id})"
        )

    def __repr__(self):
        return self.__str__()
